import com.google.common.util.concurrent.FutureCallback;
import meerkat.AsyncRegistry;
import meerkat.bulletinboard.BulletinBoardServer;
import meerkat.bulletinboard.LocalBulletinBoardClient;
import meerkat.bulletinboard.sqlserver.BulletinBoardSQLServer;
import meerkat.bulletinboard.sqlserver.H2QueryProvider;
import meerkat.comm.CommunicationException;
import meerkat.crypto.DigitalSignature;
import meerkat.crypto.concrete.ECDSASignature;
import meerkat.protobuf.BulletinBoardAPI.BulletinBoardMessage;
import meerkat.protobuf.BulletinBoardAPI.MessageFilterList;
import meerkat.protobuf.VoterRegistry;
import meerkat.protobuf.VoterRegistry.GroupID;
import meerkat.protobuf.VoterRegistry.VoterID;
import meerkat.protobuf.VoterRegistry.VoterInfo;
import meerkat.registry.MessageCollectionUtils;
import meerkat.registry.RegistryTags;
import org.junit.Before;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.KeyStore;
import java.security.SecureRandom;
import java.security.SignatureException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.Semaphore;

import static junit.framework.TestCase.assertEquals;
import static junit.framework.TestCase.assertTrue;
import static meerkat.util.BulletinBoardUtils.*;

/**
 * TODO: add logs prints for the tests to be clear what they are
 */

/**
 * Created by Vladimir Eliezer Tokarev on 1/16/2016.
 * Tests the Simple registry contents
 * NOTE: for most of this tests to pass there should run BulletinBoardServer
 * that should be reachable on BULLETIN_BOARD_SERVER_ADDRESS
 */
public class SimpleRegistryTest {

    private DigitalSignature signer;
    private LocalBulletinBoardClient localBulletinBoardClient;
    private SecureRandom random = new SecureRandom();
    Semaphore jobSemaphore;

    public static String KEYFILE_EXAMPLE = "/certs/enduser-certs/user1-key-with-password-secret.p12";
    public static String CERT1_PEM_EXAMPLE = "/certs/enduser-certs/user1.crt";
    public static String KEYFILE_PASSWORD = "secret";
    private static final String DB_NAME = "TestDB";
    private static final int SUBSCRIPTION_DELAY = 3000;
    private static final int THREAD_NUM = 3;
    final Logger logger = LoggerFactory.getLogger(getClass());

    class DummyRegistryCallBackHandler<T> implements FutureCallback<T>{
        public int counter;
        public T data;

        public DummyRegistryCallBackHandler()
        {
            counter=0;
        }

        @Override
        public void onSuccess(T result) {
            counter++;
            data = result;
            jobSemaphore.release();
        }

        @Override
        public void onFailure(Throwable throwable) {
            System.out.print(throwable.toString());
        }
    }

    public class DummyBulletinBoardCallBackHandler implements FutureCallback<List<BulletinBoardMessage>> {
        public  List<BulletinBoardMessage> messages;

        @Override
        public void onSuccess(List<BulletinBoardMessage> msg)
        {
            messages = msg;
            jobSemaphore.release();
        }

        @Override
        public void onFailure(Throwable t){
            messages = null;
            jobSemaphore.release();
        }
    }

    /**
     * Creates LocalBulletinBoardClient which connects to TestDB
     * @return LocalBulletinBoardClient object
     * @throws CommunicationException
     */
    private LocalBulletinBoardClient CreateLocalBulletinBoardClient() throws CommunicationException, SQLException {
        H2QueryProvider queryProvider = new H2QueryProvider(DB_NAME) ;

        Connection conn = queryProvider.getDataSource().getConnection();
        Statement stmt = conn.createStatement();

        List<String> deletionQueries = queryProvider.getSchemaDeletionCommands();

        for (String deletionQuery : deletionQueries) {
            stmt.execute(deletionQuery);
        }

        BulletinBoardServer server = new BulletinBoardSQLServer(queryProvider);
        server.init(DB_NAME);

        logger.debug("Created new LocalBulletinBoardClient: {}", new Date());

        return new LocalBulletinBoardClient(server, THREAD_NUM, SUBSCRIPTION_DELAY);
    }

    /**
     * Initialize CachedBulletinBoardClient client that communicates with LocalBulletinBoardClient
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws CommunicationException
     * @throws SQLException
     */
    private void CommunicatorSetup() throws InstantiationException, IllegalAccessException, CommunicationException, SQLException {
        localBulletinBoardClient = CreateLocalBulletinBoardClient();
    }

    /**
     * Sets the signer to be instance of ECDSASignature
     */
    private void SetSigner(){
        try {
            ECDSASignature signer = new ECDSASignature();
            InputStream keyStream = getClass().getResourceAsStream(KEYFILE_EXAMPLE);
            char[] password = KEYFILE_PASSWORD.toCharArray();

            KeyStore.Builder keyStore = signer.getPKCS12KeyStoreBuilder(keyStream, password);
            signer.loadSigningCertificate(keyStore);
            signer.loadVerificationCertificates(getClass().getResourceAsStream(CERT1_PEM_EXAMPLE));

            keyStream.close();
            this.signer = signer;
            logger.info("Have set new signer: {}", new Date());
        }
        catch (Exception e){
            assert false : "The signers creation failed ";
        }
    }

    /**
     * Initializes the needed params for all the tests
     * @throws SQLException
     * @throws InstantiationException
     * @throws CommunicationException
     * @throws IllegalAccessException
     */
    @Before
    public void setUp() throws SQLException, InstantiationException, CommunicationException, IllegalAccessException {
        SetSigner();
        CommunicatorSetup();
        jobSemaphore = new Semaphore(0);
        logger.info("Ended the setup preparations: {}", new Date());
    }

    private AsyncRegistry GetRegistry()
    {
        AsyncRegistry registry = new AsyncRegistry();
        registry.init(signer, localBulletinBoardClient);
        logger.info("Created new AsyncRegistry: {}", new Date());
        return registry;
    }

    /**
     * Generates random string 32 chars length
     * @return String object
     */
    private String generateString()
    {
        return new BigInteger(130, random).toString(32);
    }

    /**
     * Checks if the creation of the registry have been successful
     */
    @Test
    public void simpleRegistryCreation() {
        try {
            logger.info("Testing registry creation: {}", new Date());
            GetRegistry();
            logger.info("Registry creation succeed: {} \n\n", new Date());
        } catch (Exception e) {
            assert false : "While creating the registry exception have been thrown " + e;
        }
    }

    /**
     * Counts the amount of messages from messages that have all the wanted tags inside
     * @param messages  List<VoterRegistryMessage>
     * @param tags  List<RegistryTags>
     * @return  integer that represent the amount of messages with wanted tags
     */
    private int countMessagesWithTags(List<BulletinBoardMessage> messages, List<String> tags)
    {
        int counter = 0 ;

        for (BulletinBoardMessage message : messages) {
            int wantedTagsCounter = 0;

            for (String tag : tags) {
                if(findTagWithPrefix(message, tag)!=null){
                    wantedTagsCounter++;
                }
            }

            if(wantedTagsCounter == tags.size())
            {
                counter++;
            }
        }
        logger.info("Counted {} messages with all wanted tags from {} given messages: {}",
                    counter,
                    messages.size(),
                    new Date());
        return counter;
    }

    /**
     * Reads messages from localBulletinBoardClient by given tags and return the callback handler
     *       object
     *
     * @param tags list of strings that represents tags
     * @return DummyBulletinBoardCallBackHandler which method will be called
     */
    private DummyBulletinBoardCallBackHandler readMessagesByTags(List<String> tags)
    {
        MessageFilterList filters = MessageCollectionUtils.generateFiltersFromTags(tags, logger);
        DummyBulletinBoardCallBackHandler bulletinHandler = new DummyBulletinBoardCallBackHandler();
        localBulletinBoardClient.readMessages(filters, bulletinHandler);
        return bulletinHandler;
    }

    /**
     * Test that add voter creates new correct bulletin board message and adds the voter
     */
    @Test
    public void testAddVoter() throws InterruptedException, SignatureException {
        logger.info("Testing registry adding voter functionality: {}", new Date());

        DummyRegistryCallBackHandler<Boolean> handler = new DummyRegistryCallBackHandler<>();

        String id = generateString();
        String data = generateString();
        VoterInfo voterInfo = VoterInfo.newBuilder().setId(VoterID.newBuilder().setId(id)).setInfo(data).build();

        AsyncRegistry registry = GetRegistry();
        registry.addVoter(voterInfo, handler);
        logger.info("Tried to add next voter to registry: id: {} {}",
                      voterInfo.getId().getId(),
                      new Date());

        jobSemaphore.acquire();
        assertEquals(1, handler.counter );

        List<String> tags = new ArrayList<String>(){{ add(RegistryTags.VOTER_ENTRY_TAG);}};
        DummyBulletinBoardCallBackHandler bulletinHandler = readMessagesByTags(tags);


        jobSemaphore.acquire();
        logger.info("Got add voter messages that have been added lately to the server: {} {}",
                     GetListOfTags(bulletinHandler.messages, RegistryTags.ID_TAG),
                     new Date());

        tags.clear();
        tags.add(RegistryTags.ID_TAG + id);

        int counter = countMessagesWithTags(bulletinHandler.messages, tags);
        assert counter == 1 : "The server don't have the new user data.";
        logger.info("The server had a message with voter data: id: {} data: {} {}", id, data, new Date());
        logger.info("Ended testing registry adding voter functionality: {}\n\n", new Date());
    }

    /**
     * Test that set voted posts creates correct bulletin board message and sets that the user have been voted
     */
    @Test
    public void testSetVoted() throws InterruptedException, SignatureException {
        logger.info("Testing registry set voted functionality: {}", new Date());
        DummyRegistryCallBackHandler<Boolean> handler = new  DummyRegistryCallBackHandler<>();

        String id = generateString();
        VoterID voterInfo = VoterID.newBuilder().setId(id).build();

        AsyncRegistry registry = GetRegistry();
        registry.setVoted(voterInfo, handler);
        logger.info("Tried to set that next voter voted : id: {} {}",
                     voterInfo.getId(),
                     new Date());

        jobSemaphore.acquire();
        assertEquals(1, handler.counter );

        List<String> tags = new ArrayList<String>(){{ add(RegistryTags.VOTE_ACTION_TAG);}};
        DummyBulletinBoardCallBackHandler bulletinHandler = readMessagesByTags(tags);

        jobSemaphore.acquire();
        logger.info("Got set voted messages that have been added lately to the server: {} {} ",
                GetListOfTags(bulletinHandler.messages, RegistryTags.ID_TAG),
                new Date());

        tags.clear();
        tags.add(RegistryTags.ID_TAG + id);

        int counter = countMessagesWithTags(bulletinHandler.messages, tags);
        assert counter == 1 : "The server don't have the new user id.";
        logger.info("The server had a message with voter data: id: {} {}", id, new Date());
        logger.info("Ended testing registry set voted functionality: {}\n\n", new Date());
    }

    /**
     * Test that get groups retrieves the right groups the user are in
     */
    @Test
    public void testSetVoterGroups() throws InterruptedException, SignatureException, IOException, ClassNotFoundException {
        logger.info("Testing registry set voter groups functionality: {}", new Date());

        DummyRegistryCallBackHandler<Boolean> handler = new  DummyRegistryCallBackHandler<>();

        String voterId = generateString();
        String groupId1 = generateString();
        VoterRegistry.VoterRegistryMessage voterInfo = VoterRegistry.VoterRegistryMessage.newBuilder()
                .setVoterID(VoterID.newBuilder().setId(voterId))
                .addGroupID(GroupID.newBuilder().setId(groupId1)).build();

        AsyncRegistry registry = GetRegistry();
        registry.setVoterGroups(voterInfo, handler);
        logger.info("Tried to set voter groups : id: {} groupId: {} {}",
                voterId,
                groupId1,
                new Date());

        jobSemaphore.acquire();
        assertEquals("The callback handler hasn't been called yet", 1, handler.counter);

        List<String> tags = new ArrayList<String>(){{add(RegistryTags.ADD_TO_GROUP_TAG);
        add(RegistryTags.ID_TAG + voterId);}};
        DummyBulletinBoardCallBackHandler bulletinHandler = readMessagesByTags(tags);

        jobSemaphore.acquire();
        logger.info("Got set voter groups messages that have been added lately to the server: {} {} ",
                GetListOfTags(bulletinHandler.messages, RegistryTags.ID_TAG),
                new Date());

        BulletinBoardMessage latestMessage = getLatestMessage(bulletinHandler.messages);

        assert findTagWithPrefix(latestMessage, RegistryTags.ID_TAG).equals(voterId) :
                "The latest message recieved is not of our voter";

        logger.info("The server had a message of our voter, data: id: {} {}",
                    findTagWithPrefix(latestMessage, RegistryTags.ID_TAG), new Date());

        VoterRegistry.VoterRegistryMessage voterRegistryMessage =
                VoterRegistry.VoterRegistryMessage.parseFrom(latestMessage.getMsg().getData());

        List<GroupID> groupsIds = voterRegistryMessage.getGroupIDList();

        assert groupsIds.get(0).getId().equals(groupId1) : "The latest message doesn't have the voter group";
        logger.info("The server had a message with voter data: groupId: {} {}",
                    groupsIds.get(0).getId(), new Date());
        logger.info("Ended testing registry set voter groups functionality: {}\n\n", new Date());
    }

    /**
     * Test that remove from group creates correct bulletin board message and removes the user from a group
     */
    @Test
    public void testGetGroups() throws InterruptedException, SignatureException, IOException {
        logger.info("Testing registry get voter groups functionality: {}", new Date());
        DummyRegistryCallBackHandler<Boolean> handler =
                new  DummyRegistryCallBackHandler<>();

        String voterId = generateString();
        String groupId1 = generateString();
        String groupId2 = generateString();
        VoterRegistry.VoterRegistryMessage voterInfo = VoterRegistry.VoterRegistryMessage.newBuilder()
                .setVoterID(VoterID.newBuilder().setId(voterId))
                .addGroupID(GroupID.newBuilder().setId(groupId1))
                .addGroupID(GroupID.newBuilder().setId(groupId2)).build();

        AsyncRegistry registry = GetRegistry();
        registry.setVoterGroups(voterInfo, handler);
        logger.info("Tried to set voter groups : id: {} groupId 1: {}  groupId 2: {} {}",
                    voterId,
                    groupId1,
                    groupId2,
                    new Date());

        jobSemaphore.acquire();
        assertEquals("The callback handler hasn't been called yet", 1, handler.counter );

        DummyRegistryCallBackHandler<List<GroupID>> groupsHandler = new DummyRegistryCallBackHandler<>();
        registry.getGroups(VoterID.newBuilder().setId(voterId).build(), groupsHandler);

        jobSemaphore.acquire(1);
        List<GroupID> userGroups = groupsHandler.data;

        logger.info("Got get voter groups messages that have been added lately to the server: {}",
                new Date());

        assert userGroups.contains(GroupID.newBuilder().setId(groupId1).build()) :
                "The simple voter registry object does not retrieved the first user groups";

        assert userGroups.contains(GroupID.newBuilder().setId(groupId2).build()) :
                "The simple voter registry object does not retrieved the second user groups";

        logger.info("The server had a message with wanted data: groupId 1: {} groupId 2: {} {}",
                groupId1, groupId2, new Date());
        logger.info("Ended testing registry get voter groups functionality: {}\n\n", new Date());
    }

    /**
      * Test that the personal data outputted about the user is right
      */
    @Test
    public void testGetVoter() throws InterruptedException, SignatureException {
        logger.info("Testing registry get voter groups functionality: {}", new Date());
        DummyRegistryCallBackHandler<Boolean> handler =
                new DummyRegistryCallBackHandler<>();

        String id = generateString();
        String data = generateString();
        VoterInfo voterInfo = VoterInfo.newBuilder().
                setId(VoterID.newBuilder().setId(id)).setInfo(data).build();

        AsyncRegistry registry = GetRegistry();
        registry.addVoter(voterInfo, handler);

        logger.info("Tried to get voter with data: id: {} groupId 1: {}  groupId 2: {} {}",
                    id, data, new Date());

        jobSemaphore.acquire();
        assertEquals("The callback handler hasn't been called yet", 1, handler.counter );

        DummyRegistryCallBackHandler<VoterInfo> personalHandler = new DummyRegistryCallBackHandler<>();
        registry.getVoter(VoterID.newBuilder().setId(id).build(), personalHandler);

        jobSemaphore.acquire(1);

        logger.info("Got get voter messages that have been added lately to the server: {} {}",
                personalHandler.data.getId(),
                new Date());

        assertEquals("The voter id doesn't match the created on ",
                id, personalHandler.data.getId().getId());
        String personalData = personalHandler.data.getInfo();
        assertTrue("The voter personal data can't be found.",  data.equals(personalData));

        logger.info("The server had a message with wanted data: if: {} personal data: {} {}",
                    personalHandler.data.getId().getId(), personalData, new Date());
        logger.info("Ended testing registry get voter functionality: {}\n\n", new Date());
    }

    /**
     * Tests that the hasVoted method of registry works
     * @throws InterruptedException
     */
    @Test
    public void  testHasVoted () throws InterruptedException, SignatureException {
        logger.info("Testing registry has voted functionality: {}", new Date());

        DummyRegistryCallBackHandler<Boolean> handler = new DummyRegistryCallBackHandler<>();
        String id = generateString();
        VoterID voterInfo = VoterID.newBuilder().setId(id).build();

        AsyncRegistry registry = GetRegistry();
        registry.setVoted(voterInfo, handler);

        logger.info("Tried to set voted with data: id: {} groupId 1: {} {}",
                id, new Date());

        jobSemaphore.acquire();
        assertEquals("The callback handler hasn't been called yet", 1, handler.counter);
        DummyRegistryCallBackHandler<Boolean> personalHandler = new DummyRegistryCallBackHandler<>();
        registry.hasVoted(VoterID.newBuilder().setId(id).build(), personalHandler);

        jobSemaphore.acquire(1);

        logger.info("Got has voted messages that have been added lately to the server: {} {}",
                personalHandler.data,
                new Date());
        assertTrue("The voter hasn't voted yet.", personalHandler.data);

        logger.info("Ended testing registry has voted functionality: {}\n\n", new Date());
    }
}

